@@ -74,6 +74,12 @@ class rcube_db_sqlite extends rcube_db
                 }
             }
         }
+
+        // Enable WAL mode to fix locking issues like #8035.
+        $dbh->query("PRAGMA journal_mode = WAL");
+
+        // Enable foreign keys (requires sqlite 3.6.19 compiled with FK support)
+        $dbh->query("PRAGMA foreign_keys = ON");
     }
 
     /**
@@ -145,4 +151,21 @@ class rcube_db_sqlite extends rcube_db
     {
         return $dsn['phptype'] . ':' . $dsn['database'];
     }
+
+    /**
+     * Returns driver-specific connection options
+     *
+     * @param array $dsn DSN parameters
+     *
+     * @return array Connection options
+     */
+    protected function dsn_options($dsn)
+    {
+        $result = parent::dsn_options($dsn);
+
+        // Change the default timeout (60) to a smaller value
+        $result[PDO::ATTR_TIMEOUT] = isset($dsn['timeout']) ? intval($dsn['timeout']) : 10;
+
+        return $result;
+    }
 }
