@@ -98,6 +98,11 @@
  */
 class rcube_html2text
 {
+    const LINKS_NONE = 0;
+    const LINKS_END = 1;
+    const LINKS_INLINE = 2;
+    const LINKS_DEFAULT = self::LINKS_END;
+
     /**
      * Contains the HTML content to convert.
      *
@@ -118,7 +123,7 @@ class rcube_html2text
      * Set this value to 0 (or less) to ignore word wrapping
      * and not constrain text to a fixed-width column.
      *
-     * @var integer $width
+     * @var int $width
      */
     protected $width = 70;
 
@@ -134,13 +139,14 @@ class rcube_html2text
      * used in conjunction with $replace.
      *
      * @var array $search
-     * @see $replace
+     * @see self::$replace
      */
     protected $search = [
         '/\r/',                                  // Non-legal carriage return
-        '/<head[^>]*>.*?<\/head>/is',            // <head>
-        '/<script[^>]*>.*?<\/script>/is',        // <script>
-        '/<style[^>]*>.*?<\/style>/is',          // <style>
+        '/\n*<\/?html>\n*/is',                   // <html>
+        '/\n*<head[^>]*>.*?<\/head>\n*/is',      // <head>
+        '/\n*<script[^>]*>.*?<\/script>\n*/is',  // <script>
+        '/\n*<style[^>]*>.*?<\/style>\n*/is',    // <style>
         '/[\n\t]+/',                             // Newlines and tabs
         '/<p[^>]*>/i',                           // <p>
         '/<\/p>[\s\n\t]*<div[^>]*>/i',           // </p> before <div>
@@ -163,10 +169,11 @@ class rcube_html2text
      * List of pattern replacements corresponding to patterns searched.
      *
      * @var array $replace
-     * @see $search
+     * @see self::$search
      */
     protected $replace = [
         '',                                     // Non-legal carriage return
+        '',                                     // <html>|</html>
         '',                                     // <head>
         '',                                     // <script>
         '',                                     // <style>
@@ -193,7 +200,7 @@ class rcube_html2text
      * used in conjunction with $ent_replace.
      *
      * @var array $ent_search
-     * @see $ent_replace
+     * @see self::$ent_replace
      */
     protected $ent_search = [
         '/&(nbsp|#160);/i',                      // Non-breaking space
@@ -217,7 +224,7 @@ class rcube_html2text
      * List of pattern replacements corresponding to patterns searched.
      *
      * @var array $ent_replace
-     * @see $ent_search
+     * @see self::$ent_search
      */
     protected $ent_replace = [
         "\xC2\xA0",                             // Non-breaking space
@@ -254,7 +261,7 @@ class rcube_html2text
      * used in conjunction with $pre_replace.
      *
      * @var array $pre_search
-     * @see $pre_replace
+     * @see self::$pre_replace
      */
     protected $pre_search = [
         "/\n/",
@@ -268,7 +275,7 @@ class rcube_html2text
      * List of pattern replacements corresponding to patterns searched for PRE body.
      *
      * @var array $pre_replace
-     * @see $pre_search
+     * @see self::$pre_search
      */
     protected $pre_replace = [
         '<br>',
@@ -289,7 +296,7 @@ class rcube_html2text
      * Contains a list of HTML tags to allow in the resulting text.
      *
      * @var string $allowed_tags
-     * @see set_allowed_tags()
+     * @see self::set_allowed_tags()
      */
     protected $allowed_tags = '';
 
@@ -303,8 +310,9 @@ class rcube_html2text
     /**
      * Indicates whether content in the $html variable has been converted yet.
      *
-     * @var boolean $_converted
-     * @see $html, $text
+     * @var bool $_converted
+     * @see self::$html
+     * @see self::$text
      */
     protected $_converted = false;
 
@@ -312,17 +320,19 @@ class rcube_html2text
      * Contains URL addresses from links to be rendered in plain text.
      *
      * @var array $_link_list
-     * @see _build_link_list()
+     * @see self::_build_link_list()
      */
     protected $_link_list = [];
 
     /**
-     * Boolean flag, true if a table of link URLs should be listed after the text.
+     * Links handling.
+     * - 0 if links should be removed
+     * - 1 if a table of link URLs should be listed after the text
+     * - 2 if the link should be displayed to the original point in the text they appeared
      *
-     * @var boolean $_do_links
-     * @see __construct()
+     * @var int $_links_mode
      */
-    protected $_do_links = true;
+    protected $_links_mode = 1;
 
     /**
      * Constructor.
@@ -331,22 +341,43 @@ class rcube_html2text
      * will instantiate with that source propagated, all that has
      * to be done it to call get_text().
      *
-     * @param string $source    HTML content
-     * @param bool   $from_file Indicates $source is a file to pull content from
-     * @param bool   $do_links  Indicate whether a table of link URLs is desired
-     * @param int    $width     Maximum width of the formatted text, 0 for no limit
+     * @param string   $source     HTML content
+     * @param bool     $from_file  Indicates $source is a file to pull content from
+     * @param bool|int $links_mode Links handling mode
+     * @param int      $width      Maximum width of the formatted text, 0 for no limit
      */
-    function __construct($source = '', $from_file = false, $do_links = true, $width = 75, $charset = 'UTF-8')
+    function __construct($source = '', $from_file = false, $links_mode = self::LINKS_DEFAULT, $width = 75, $charset = 'UTF-8')
     {
         if (!empty($source)) {
             $this->set_html($source, $from_file);
         }
 
         $this->set_base_url();
+        $this->set_links_mode($links_mode);
 
-        $this->_do_links = $do_links;
-        $this->width     = $width;
-        $this->charset   = $charset;
+        $this->width   = $width;
+        $this->charset = $charset;
+    }
+
+    /**
+     * Sets the links behavior mode
+     *
+     * @param bool|int $mode
+     */
+    private function set_links_mode($mode)
+    {
+        $allowed = [
+            self::LINKS_NONE,
+            self::LINKS_END,
+            self::LINKS_INLINE
+        ];
+
+        if (!in_array((int) $mode, $allowed)) {
+            $this->_links_mode = self::LINKS_DEFAULT;
+            return;
+        }
+
+        $this->_links_mode = (int) $mode;
     }
 
     /**
@@ -473,6 +504,7 @@ class rcube_html2text
         if (($pos = stripos($text, '<body')) !== false) {
             $pos = strpos($text, '>', $pos);
             $text = substr($text, $pos + 1);
+            $text = ltrim($text);
         }
 
         // Run our defined tags search-and-replace
@@ -519,14 +551,14 @@ class rcube_html2text
      * Helper function called by preg_replace() on link replacement.
      *
      * Maintains an internal list of links to be displayed at the end of the
-     * text, with numeric indices to the original point in the text they
+     * text, with numeric indices or simply the link to the original point in the text they
      * appeared. Also makes an effort at identifying and handling absolute
      * and relative links.
      *
      * @param string $link    URL of the link
      * @param string $display Part of the text to associate number with
      */
-    protected function _build_link_list($link, $display)
+    protected function _handle_link($link, $display)
     {
         if (empty($link)) {
             return $display;
@@ -553,7 +585,7 @@ class rcube_html2text
             $url .= "$link";
         }
 
-        if (!$this->_do_links) {
+        if (self::LINKS_NONE === $this->_links_mode) {
             // When not using link list use URL if there's no content (#5795)
             // The content here is HTML, convert it to text first
             $h2t     = new rcube_html2text($display, false, false, 1024, $this->charset);
@@ -566,6 +598,39 @@ class rcube_html2text
             return $display;
         }
 
+        if (self::LINKS_INLINE === $this->_links_mode) {
+            return $this->_build_link_inline($url, $display);
+        }
+
+        return $this->_build_link_list($url, $display);
+    }
+
+    /**
+     * Helper function called by _handle_link() on link replacement.
+     *
+     * Displays the link next to the original point in the text they
+     * appeared.
+     *
+     * @param string $url     URL of the link
+     * @param string $display linktext
+     */
+    protected function _build_link_inline($url, $display)
+    {
+        return $display . ' &lt;' . $url . '&gt;';
+    }
+
+    /**
+     * Helper function called by _handle_link() on link replacement.
+     *
+     * Maintains an internal list of links to be displayed at the end of the
+     * text, with numeric indices to the original point in the text they
+     * appeared.
+     *
+     * @param string $url    URL of the link
+     * @param string $display Part of the text to associate number with
+     */
+    protected function _build_link_list($url, $display)
+    {
         if (($index = array_search($url, $this->_link_list)) === false) {
             $index = count($this->_link_list);
             $this->_link_list[] = $url;
@@ -693,7 +758,8 @@ class rcube_html2text
         case 'a':
             // Remove spaces in URL (#1487805)
             $url = str_replace(' ', '', $matches[3]);
-            return $this->_build_link_list($url, $matches[4]);
+            $url = html_entity_decode($url, \ENT_HTML5, RCUBE_CHARSET);
+            return $this->_handle_link($url, $matches[4]);
         }
     }
 
@@ -719,7 +785,7 @@ class rcube_html2text
     private function _toupper($str)
     {
         // string can containing HTML tags
-        $chunks = preg_split('/(<[^>]*>)/', $str, null, PREG_SPLIT_NO_EMPTY | PREG_SPLIT_DELIM_CAPTURE);
+        $chunks = preg_split('/(<[^>]*>)/', $str, -1, PREG_SPLIT_NO_EMPTY | PREG_SPLIT_DELIM_CAPTURE);
 
         // convert toupper only the text between HTML tags
         foreach ($chunks as $idx => $chunk) {
